@@ -45,19 +45,24 @@
 #include <rpc/xdr.h>
 #include <rpc/auth.h>
 #include <rpc/auth_gss.h>
+#include <rpc/rpcsec_gss.h>
 #include <rpc/clnt.h>
 #include <netinet/in.h>
-#include <gssapi/gssapi.h>
 
-static void	authgss_nextverf();
-static bool_t	authgss_marshal();
-static bool_t	authgss_refresh();
-static bool_t	authgss_validate();
-static void	authgss_destroy();
-static void	authgss_destroy_context();
-static bool_t	authgss_wrap();
-static bool_t	authgss_unwrap();
+#include "debug.h"
 
+static void	authgss_nextverf(AUTH *);
+static bool_t	authgss_marshal(AUTH *, XDR *);
+static bool_t	authgss_refresh(AUTH *, void *);
+static bool_t	authgss_validate(AUTH *, struct opaque_auth *);
+static void	authgss_destroy(AUTH *);
+static void	authgss_destroy_context(AUTH *);
+static bool_t	authgss_wrap(AUTH *, XDR *, xdrproc_t, caddr_t);
+static bool_t	authgss_unwrap(AUTH *, XDR *, xdrproc_t, caddr_t);
+
+void rpc_gss_set_error(int);
+void rpc_gss_clear_error(void);
+bool_t rpc_gss_oid_to_mech(rpc_gss_OID, char **);
 
 /*
  * from mit-krb5-1.2.1 mechglue/mglueP.h:
@@ -78,7 +83,6 @@ static struct auth_ops authgss_ops = {
 	authgss_unwrap
 };
 
-#ifdef DEBUG
 
 /* useful as i add more mechanisms */
 void
@@ -87,6 +91,9 @@ print_rpc_gss_sec(struct rpc_gss_sec *ptr)
 int i;
 char *p;
 
+	if (libtirpc_debug_level < 4 || log_stderr == 0)
+		return;
+
 	gss_log_debug("rpc_gss_sec:");
 	if(ptr->mech == NULL)
 		gss_log_debug("NULL gss_OID mech");
@@ -126,7 +133,6 @@ char *p;
 	fprintf(stderr, "     service: %d\n", ptr->svc);
 	fprintf(stderr, "     cred: %p\n", ptr->cred);
 }
-#endif /*DEBUG*/
 
 struct rpc_gss_data {
 	bool_t			 established;	/* context established */
@@ -166,9 +172,7 @@ authgss_create(CLIENT *clnt, gss_name_t name, struct rpc_gss_sec *sec)
 		free(auth);
 		return (NULL);
 	}
-#ifdef DEBUG
-	fprintf(stderr, "authgss_create: name is %p\n", name);
-#endif
+	LIBTIRPC_DEBUG(3, ("authgss_create: name is %p", name));
 	if (name != GSS_C_NO_NAME) {
 		if (gss_duplicate_name(&min_stat, name, &gd->name)
 						!= GSS_S_COMPLETE) {
@@ -181,9 +185,7 @@ authgss_create(CLIENT *clnt, gss_name_t name, struct rpc_gss_sec *sec)
 	else
 		gd->name = name;
 
-#ifdef DEBUG
-	fprintf(stderr, "authgss_create: gd->name is %p\n", gd->name);
-#endif
+	LIBTIRPC_DEBUG(3, ("authgss_create: gd->name is %p", gd->name));
 	gd->clnt = clnt;
 	gd->ctx = GSS_C_NO_CONTEXT;
 	gd->sec = *sec;
@@ -198,7 +200,7 @@ authgss_create(CLIENT *clnt, gss_name_t name, struct rpc_gss_sec *sec)
 	save_auth = clnt->cl_auth;
 	clnt->cl_auth = auth;
 
-	if (!authgss_refresh(auth))
+	if (!authgss_refresh(auth, NULL))
 		auth = NULL;
 	else
 		auth_get(auth); /* Reference for caller */
@@ -227,7 +229,8 @@ authgss_create_default(CLIENT *clnt, char *service, struct rpc_gss_sec *sec)
 		&name);
 
 	if (maj_stat != GSS_S_COMPLETE) {
-		gss_log_status("gss_import_name", maj_stat, min_stat);
+		gss_log_status("authgss_create_default: gss_import_name", 
+			maj_stat, min_stat);
 		rpc_createerr.cf_stat = RPC_AUTHERROR;
 		return (NULL);
 	}
@@ -235,9 +238,7 @@ authgss_create_default(CLIENT *clnt, char *service, struct rpc_gss_sec *sec)
 	auth = authgss_create(clnt, name, sec);
 
 	if (name != GSS_C_NO_NAME) {
-#ifdef DEBUG
-	fprintf(stderr, "authgss_create_default: freeing name %p\n", name);
-#endif
+		LIBTIRPC_DEBUG(3, ("authgss_create_default: freeing name %p", name));
  		gss_release_name(&min_stat, &name);
 	}
 
@@ -346,7 +347,8 @@ authgss_marshal(AUTH *auth, XDR *xdrs)
 			    &rpcbuf, &checksum);
 
 	if (maj_stat != GSS_S_COMPLETE) {
-		gss_log_status("gss_get_mic", maj_stat, min_stat);
+		gss_log_status("authgss_marshal: gss_get_mic", 
+			maj_stat, min_stat);
 		if (maj_stat == GSS_S_CONTEXT_EXPIRED) {
 			gd->established = FALSE;
 			authgss_destroy_context(auth);
@@ -406,8 +408,10 @@ authgss_validate(AUTH *auth, struct opaque_auth *verf)
 
 	maj_stat = gss_verify_mic(&min_stat, gd->ctx, &signbuf,
 				  &checksum, &qop_state);
+
 	if (maj_stat != GSS_S_COMPLETE || qop_state != gd->sec.qop) {
-		gss_log_status("gss_verify_mic", maj_stat, min_stat);
+		gss_log_status("authgss_validate: gss_verify_mic", 
+			maj_stat, min_stat);
 		if (maj_stat == GSS_S_CONTEXT_EXPIRED) {
 			gd->established = FALSE;
 			authgss_destroy_context(auth);
@@ -418,12 +422,15 @@ authgss_validate(AUTH *auth, struct opaque_auth *verf)
 }
 
 static bool_t
-authgss_refresh(AUTH *auth)
+_rpc_gss_refresh(AUTH *auth, rpc_gss_options_ret_t *options_ret)
 {
 	struct rpc_gss_data	*gd;
 	struct rpc_gss_init_res	 gr;
 	gss_buffer_desc		*recv_tokenp, send_token;
-	OM_uint32		 maj_stat, min_stat, call_stat, ret_flags;
+	OM_uint32		 maj_stat, min_stat, call_stat, ret_flags,
+				 time_ret;
+	gss_OID			 actual_mech_type;
+	char			*mechanism;
 
 	gss_log_debug("in authgss_refresh()");
 
@@ -436,19 +443,15 @@ authgss_refresh(AUTH *auth)
 	memset(&gr, 0, sizeof(gr));
 	recv_tokenp = GSS_C_NO_BUFFER;
 
-#ifdef DEBUG
 	print_rpc_gss_sec(&gd->sec);
-#endif /*DEBUG*/
 
 	for (;;) {
-#ifdef DEBUG
 		/* print the token we just received */
 		if (recv_tokenp != GSS_C_NO_BUFFER) {
 			gss_log_debug("The token we just received (length %d):",
 				  recv_tokenp->length);
 			gss_log_hexdump(recv_tokenp->value, recv_tokenp->length, 0);
 		}
-#endif
 		maj_stat = gss_init_sec_context(&min_stat,
 						gd->sec.cred,
 						&gd->ctx,
@@ -458,10 +461,10 @@ authgss_refresh(AUTH *auth)
 						0,		/* time req */
 						NULL,		/* channel */
 						recv_tokenp,
-						NULL,		/* used mech */
+						&actual_mech_type,
 						&send_token,
 						&ret_flags,
-						NULL);		/* time rec */
+						&time_ret);
 
 		if (recv_tokenp != GSS_C_NO_BUFFER) {
 			gss_release_buffer(&min_stat, &gr.gr_token);
@@ -470,17 +473,17 @@ authgss_refresh(AUTH *auth)
 		if (maj_stat != GSS_S_COMPLETE &&
 		    maj_stat != GSS_S_CONTINUE_NEEDED) {
 			gss_log_status("gss_init_sec_context", maj_stat, min_stat);
+			options_ret->major_status = maj_stat;
+			options_ret->minor_status = min_stat;
 			break;
 		}
 		if (send_token.length != 0) {
 			memset(&gr, 0, sizeof(gr));
 
-#ifdef DEBUG
 			/* print the token we are about to send */
 			gss_log_debug("The token being sent (length %d):",
 				  send_token.length);
 			gss_log_hexdump(send_token.value, send_token.length, 0);
-#endif
 
 			call_stat = clnt_call(gd->clnt, NULLPROC,
 					      (xdrproc_t)xdr_rpc_gss_init_args,
@@ -492,8 +495,11 @@ authgss_refresh(AUTH *auth)
 
 			if (call_stat != RPC_SUCCESS ||
 			    (gr.gr_major != GSS_S_COMPLETE &&
-			     gr.gr_major != GSS_S_CONTINUE_NEEDED))
+			     gr.gr_major != GSS_S_CONTINUE_NEEDED)) {
+				options_ret->major_status = gr.gr_major;
+				options_ret->minor_status = gr.gr_minor;
 				return FALSE;
+			}
 
 			if (gr.gr_ctx.length != 0) {
 				if (gd->gc.gc_ctx.value)
@@ -528,13 +534,31 @@ authgss_refresh(AUTH *auth)
 
 			if (maj_stat != GSS_S_COMPLETE
 					|| qop_state != gd->sec.qop) {
-				gss_log_status("gss_verify_mic", maj_stat, min_stat);
+				gss_log_status("authgss_refresh: gss_verify_mic", 
+					maj_stat, min_stat);
 				if (maj_stat == GSS_S_CONTEXT_EXPIRED) {
 					gd->established = FALSE;
 					authgss_destroy_context(auth);
 				}
+				rpc_gss_set_error(EPERM);
+				options_ret->major_status = maj_stat;
+				options_ret->minor_status = min_stat;
 				return (FALSE);
 			}
+
+			options_ret->major_status = GSS_S_COMPLETE;
+			options_ret->minor_status = 0;
+			options_ret->rpcsec_version = gd->gc.gc_v;
+			options_ret->ret_flags = ret_flags;
+			options_ret->time_ret = time_ret;
+			options_ret->gss_context = gd->ctx;
+			options_ret->actual_mechanism[0] = '\0';
+			if (rpc_gss_oid_to_mech(actual_mech_type, &mechanism)) {
+				strncpy(options_ret->actual_mechanism,
+					mechanism,
+					sizeof(options_ret->actual_mechanism));
+			}
+
 			gd->established = TRUE;
 			gd->gc.gc_proc = RPCSEC_GSS_DATA;
 			gd->gc.gc_seq = 0;
@@ -550,12 +574,22 @@ authgss_refresh(AUTH *auth)
 		authgss_destroy(auth);
 		auth = NULL;
 		rpc_createerr.cf_stat = RPC_AUTHERROR;
+		rpc_gss_set_error(EPERM);
 
 		return (FALSE);
 	}
 	return (TRUE);
 }
 
+static bool_t
+authgss_refresh(AUTH *auth, void *dummy)
+{
+	rpc_gss_options_ret_t ret;
+
+	memset(&ret, 0, sizeof(ret));
+	return _rpc_gss_refresh(auth, &ret);
+}
+
 bool_t
 authgss_service(AUTH *auth, int svc)
 {
@@ -629,9 +663,7 @@ authgss_destroy(AUTH *auth)
 
 	authgss_destroy_context(auth);
 
-#ifdef DEBUG
-	fprintf(stderr, "authgss_destroy: freeing name %p\n", gd->name);
-#endif
+	LIBTIRPC_DEBUG(3, ("authgss_destroy: freeing name %p", gd->name));
 	if (gd->name != GSS_C_NO_NAME)
 		gss_release_name(&min_stat, &gd->name);
 
@@ -639,7 +671,7 @@ authgss_destroy(AUTH *auth)
 	free(auth);
 }
 
-bool_t
+static bool_t
 authgss_wrap(AUTH *auth, XDR *xdrs, xdrproc_t xdr_func, caddr_t xdr_ptr)
 {
 	struct rpc_gss_data	*gd;
@@ -656,7 +688,7 @@ authgss_wrap(AUTH *auth, XDR *xdrs, xdrproc_t xdr_func, caddr_t xdr_ptr)
 				 gd->sec.svc, gd->gc.gc_seq));
 }
 
-bool_t
+static bool_t
 authgss_unwrap(AUTH *auth, XDR *xdrs, xdrproc_t xdr_func, caddr_t xdr_ptr)
 {
 	struct rpc_gss_data	*gd;
@@ -672,3 +704,240 @@ authgss_unwrap(AUTH *auth, XDR *xdrs, xdrproc_t xdr_func, caddr_t xdr_ptr)
 				 gd->ctx, gd->sec.qop,
 				 gd->sec.svc, gd->gc.gc_seq));
 }
+
+static AUTH *
+_rpc_gss_seccreate_error(int system_error)
+{
+	rpc_gss_set_error(system_error);
+	return NULL;
+}
+
+/*
+ * External API: Create a GSS security context for this "clnt"
+ *
+ * clnt: a valid RPC CLIENT structure
+ * principal: NUL-terminated C string containing GSS principal
+ * mechanism: NUL-terminated C string containing GSS mechanism name
+ * service: GSS security value to use
+ * qop: NUL-terminated C string containing QOP name
+ * req: pointer to additional request parameters, or NULL
+ * ret: pointer to additional results, or NULL
+ *
+ * Returns pointer to a filled in RPC AUTH structure or NULL.
+ * Caller must free returned structure with AUTH_DESTROY.
+ */
+AUTH *
+rpc_gss_seccreate(CLIENT *clnt, char *principal, char *mechanism,
+		rpc_gss_service_t service, char *qop,
+		rpc_gss_options_req_t *req, rpc_gss_options_ret_t *ret)
+{
+	rpc_gss_options_ret_t options_ret;
+	OM_uint32 maj_stat, min_stat;
+	struct rpc_gss_sec sec = {
+		.req_flags	= GSS_C_MUTUAL_FLAG,
+		.cred		= GSS_C_NO_CREDENTIAL,
+	};
+	struct rpc_gss_data *gd;
+	AUTH *auth, *save_auth;
+	gss_buffer_desc sname;
+
+	if (clnt == NULL || principal == NULL)
+		return _rpc_gss_seccreate_error(EINVAL);
+
+	if (rpc_gss_mech_to_oid(mechanism, &sec.mech) == FALSE)
+		return NULL;
+
+	sec.qop = GSS_C_QOP_DEFAULT;
+	if (qop != NULL) {
+		u_int qop_num;
+		if (rpc_gss_qop_to_num(qop, mechanism, &qop_num) == FALSE)
+			return NULL;
+		sec.qop = qop_num;
+	}
+
+	switch (service) {
+	case rpcsec_gss_svc_none:
+		sec.svc = RPCSEC_GSS_SVC_NONE;
+		break;
+	case rpcsec_gss_svc_integrity:
+		sec.svc = RPCSEC_GSS_SVC_INTEGRITY;
+		break;
+	case rpcsec_gss_svc_privacy:
+		sec.svc = RPCSEC_GSS_SVC_PRIVACY;
+		break;
+	case rpcsec_gss_svc_default:
+		sec.svc = RPCSEC_GSS_SVC_INTEGRITY;
+		break;
+	default:
+		return _rpc_gss_seccreate_error(ENOENT);
+	}
+
+	if (req != NULL) {
+		sec.req_flags = req->req_flags;
+		sec.cred = req->my_cred;
+	}
+
+	if (ret == NULL)
+		ret = &options_ret;
+	memset(ret, 0, sizeof(*ret));
+
+	auth = calloc(1, sizeof(*auth));
+	gd = calloc(1, sizeof(*gd));
+	if (auth == NULL || gd == NULL) {
+		free(gd);
+		free(auth);
+		return _rpc_gss_seccreate_error(ENOMEM);
+	}
+
+	sname.value = principal;
+	sname.length = strlen(principal);
+	maj_stat = gss_import_name(&min_stat, &sname,
+					(gss_OID)GSS_C_NT_HOSTBASED_SERVICE,
+					&gd->name);
+	if (maj_stat != GSS_S_COMPLETE) {
+		ret->major_status = maj_stat;
+		ret->minor_status = min_stat;
+		free(gd);
+		free(auth);
+		return _rpc_gss_seccreate_error(ENOMEM);
+	}
+
+	gd->clnt = clnt;
+	gd->ctx = GSS_C_NO_CONTEXT;
+	gd->sec = sec;
+
+	gd->gc.gc_v = RPCSEC_GSS_VERSION;
+	gd->gc.gc_proc = RPCSEC_GSS_INIT;
+	gd->gc.gc_svc = sec.svc;
+
+	auth->ah_ops = &authgss_ops;
+	auth->ah_private = (caddr_t)gd;
+
+	save_auth = clnt->cl_auth;
+	clnt->cl_auth = auth;
+
+	if (_rpc_gss_refresh(auth, ret) == FALSE) {
+		authgss_destroy(auth);
+		auth = NULL;
+	} else {
+		rpc_gss_clear_error();
+		auth_get(auth);
+	}
+
+	clnt->cl_auth = save_auth;
+
+	return auth;
+}
+
+/*
+ * External API: Modify an AUTH's service and QOP settings
+ *
+ * auth: a valid RPC AUTH structure
+ * service: GSS security value to use
+ * qop: NUL-terminated C string containing QOP name
+ *
+ * Returns TRUE if successful, otherwise FALSE is returned.
+ */
+bool_t
+rpc_gss_set_defaults(AUTH *auth, rpc_gss_service_t service, char *qop)
+{
+	struct rpc_gss_data *gd;
+	char *mechanism;
+	u_int qop_num;
+
+	if (auth == NULL) {
+		rpc_gss_set_error(EINVAL);
+		return FALSE;
+	}
+
+	gd = AUTH_PRIVATE(auth);
+	if (gd == NULL) {
+		rpc_gss_set_error(EINVAL);
+		return FALSE;
+	}
+
+	if (rpc_gss_oid_to_mech(gd->sec.mech, &mechanism) == FALSE)
+		return FALSE;
+
+	qop_num = GSS_C_QOP_DEFAULT;
+	if (qop != NULL) {
+		if (rpc_gss_qop_to_num(qop, mechanism, &qop_num) == FALSE)
+			return FALSE;
+	}
+
+	switch (service) {
+	case rpcsec_gss_svc_none:
+		gd->gc.gc_svc = gd->sec.svc = RPCSEC_GSS_SVC_NONE;
+		break;
+	case rpcsec_gss_svc_integrity:
+		gd->gc.gc_svc = gd->sec.svc = RPCSEC_GSS_SVC_INTEGRITY;
+		break;
+	case rpcsec_gss_svc_privacy:
+		gd->gc.gc_svc = gd->sec.svc = RPCSEC_GSS_SVC_PRIVACY;
+		break;
+	case rpcsec_gss_svc_default:
+		gd->gc.gc_svc = gd->sec.svc = RPCSEC_GSS_SVC_INTEGRITY;
+		break;
+	default:
+		rpc_gss_set_error(ENOENT);
+		return FALSE;
+	}
+
+	gd->sec.qop = (gss_qop_t)qop_num;
+	rpc_gss_clear_error();
+	return TRUE;
+}
+
+/*
+ * External API: Return maximum data size for a security mechanism and transport
+ *
+ * auth: a valid RPC AUTH structure
+ * maxlen: transport's maximum data size, in bytes
+ *
+ * Returns maximum data size given transformations done by current
+ * security setting of "auth", in bytes, or zero if that value
+ * cannot be determined.
+ */
+int
+rpc_gss_max_data_length(AUTH *auth, int maxlen)
+{
+	OM_uint32 max_input_size, maj_stat, min_stat;
+	struct rpc_gss_data *gd;
+	int conf_req_flag;
+	int result;
+
+	if (auth == NULL) {
+		rpc_gss_set_error(EINVAL);
+		return 0;
+	}
+
+	gd = AUTH_PRIVATE(auth);
+	if (gd == NULL) {
+		rpc_gss_set_error(EINVAL);
+		return 0;
+	}
+
+	switch (gd->sec.svc) {
+	case RPCSEC_GSS_SVC_NONE:
+		rpc_gss_clear_error();
+		return maxlen;
+	case RPCSEC_GSS_SVC_INTEGRITY:
+		conf_req_flag = 0;
+		break;
+	case RPCSEC_GSS_SVC_PRIVACY:
+		conf_req_flag = 1;
+		break;
+	default:
+		rpc_gss_set_error(ENOENT);
+		return 0;
+	}
+
+	result = 0;
+	maj_stat = gss_wrap_size_limit(&min_stat, gd->ctx, conf_req_flag,
+					gd->sec.qop, maxlen, &max_input_size);
+	if (maj_stat == GSS_S_COMPLETE)
+		if ((int)max_input_size > 0)
+			result = (int)max_input_size;
+	rpc_gss_clear_error();
+	return result;
+}
